//=============================================================================
// TeamGamePlus.
//=============================================================================
class TeamGamePlus expands BotPack.TeamGamePlus
	config;

var Actor CaptActor;
var string MapMessage;
var bool bUseXC;
var config bool bExtraHit;

event PreBeginPlay()
{
//	SetGameSpeed(GameSpeed); //Removed
	Level.bNoCheating = bNoCheating;
	Level.bAllowFOV = bAllowFOV;

	if (GameReplicationInfoClass != None)
		GameReplicationInfo = Spawn(GameReplicationInfoClass);
	else
		GameReplicationInfo = Spawn(class'GameReplicationInfo');
	InitGameReplicationInfo();
}

event PostBeginPlay()
{
	local CaptureActor Ac;

	foreach AllActors (class 'CaptureActor',Ac)
	{
		if ( Ac != None )
		{
			CaptActor = Ac;
			log ("Game Core found CaptureActor.");
			break;
		}
	}
	Super.PostBeginPlay();
	if ( NsDeathMatchMutator(BaseMutator) != None && NsDeathMatchMutator(BaseMutator).bHaveXC )
		bUseXC = True;
}

function SetGameSpeed(Float T)
{
	local AlignSPP Ali;

	GameSpeed=FMax(T,0.1);
	Level.TimeDilation=GameSpeed;
	SetTimer(Level.TimeDilation, True);
//	SaveConfig(); //Nah, won't screw initial setup any so called Admin
	if ( Level.NetMode != NM_StandAlone ) //Track speed changing
	{
		foreach AllActors (class'AlignSPP', Ali)
		{
			Ali.Speed = GameSpeed; //All aligners from clients will get authority data
		}
	}
}

event InitGame( string Options, out string Error )
{
	local string InOpt;

	SolveWepsy();
	StartTime = 0;
	RemainingTime = 60 * TimeLimit;
	SetGameSpeed(GameSpeed);
	Super(GameInfo).InitGame(Options, Error);
	//DeathMatchPlus
	FragLimit = GetIntOption( Options, "FragLimit", FragLimit );
	TimeLimit = GetIntOption( Options, "TimeLimit", TimeLimit );
	MaxCommanders = GetIntOption( Options, "MaxCommanders", MaxCommanders );

	InOpt = ParseOption( Options, "CoopWeaponMode");
	if ( InOpt != "" )
	{
		log("CoopWeaponMode: "$bool(InOpt));
		bCoopWeaponMode = bool(InOpt);
	}

	IDnum = -1;
	IDnum = GetIntOption( Options, "Tournament", IDnum );
	if ( IDnum > 0 )
	{
		bRatedGame = true;
		TimeLimit = 0;
		RemainingTime = 0;
	}
	if ( bTournament ) 
	{
		bRequireReady = true;
		CheckReady();
	}
	if ( Level.NetMode == NM_StandAlone )
	{
		bRequireReady = true;
		CountDown = 1;
	}
	if ( !bRequireReady && (Level.NetMode != NM_Standalone) )
	{
		bRequireReady = true;
		bNetReady = true;
	}

	bNoMonsters = False;
	bWorldLog = False;
	//Craps removed, time to play
	MaxTeams = Min(MaxTeams,MaxAllowedTeams);
}

function SolveWepsy()
{
	local Inventory I;
	local string Sc;

	foreach AllActors(class'Inventory',I)
	{
		if ( I != None )
		{
			if (I.bStatic)
			{
				log(I@"is wrong set bStatic...");
				Sc = Caps(I.Class);
				Sc = "set"@Sc@"bStatic 0";
				ConsoleCommand(Sc);
//				Sc = "set"@Sc@"bRotatingPickup 0";
//				ConsoleCommand(Sc);
				I.bRotatingPickup=False;
				log (I@"has now bStatic ="@I.bStatic);
				Sc = "";
			}
			if (I.bNoDelete)
			{
				log(I@"is wrong set bNoDelete...");
				Sc = Caps(I.Class);
				Sc = "set"@Sc@"bNoDelete 0";
				ConsoleCommand(Sc);
//				Sc = "set"@Sc@"bRotatingPickup 0";
//				ConsoleCommand(Sc);
				I.bRotatingPickup=False;
				log (I@"has now bNoDelete ="@I.bNoDelete);
				Sc = "";
			}
		}
	}
	ConsoleCommand("set inventory bGameRelevant 0"); //Suck well
	ConsoleCommand("set inventory bMovable 1");
	ConsoleCommand("set inventory bTravel 1");
	ConsoleCommand("set inventory bIsItemGoal 1");
}

// Set game settings based on ladder information.
// Called when RatedPlayer logs in.

function bool SetEndCams(string Reason)
{
	local TeamInfo BestTeam;
	local int i;
	local pawn P, Best;
	local PlayerPawn player;

	// find individual winner
	for ( P=Level.PawnList; P!=None; P=P.nextPawn )
		if ( P.bIsPlayer && P.PlayerReplicationInfo != None && ((Best == None) || (P.PlayerReplicationInfo.Score > Best.PlayerReplicationInfo.Score)) )
			Best = P;

	// find winner
	BestTeam = Teams[0];
	for ( i=1; i<MaxTeams; i++ )
		if ( Teams[i].Score > BestTeam.Score )
			BestTeam = Teams[i];

	for ( i=0; i<MaxTeams; i++ )
		if ( (BestTeam.TeamIndex != i) && (BestTeam.Score == Teams[i].Score) )
		{
			BroadcastLocalizedMessage( DMMessageClass, 0 );
			return false;
		}

	GameReplicationInfo.GameEndedComments = TeamPrefix@BestTeam.TeamName@GameEndedMessage;

	EndTime = Level.TimeSeconds + 3.0;
	for ( P=Level.PawnList; P!=None; P=P.nextPawn )
	{
		player = PlayerPawn(P);
		if ( Player != None )
		{
			if (!bTutorialGame)
				PlayWinMessage(Player, (Player.PlayerReplicationInfo.Team == BestTeam.TeamIndex));
			player.bBehindView = true;
			if ( Player == Best )
				Player.ViewTarget = None;
			else
				Player.ViewTarget = Best;
			player.ClientGameEnded();
		}
		P.GotoState('GameEnded');
	}
//	CalcEndStats();
	return true;
}

//------------------------------------------------------------------------------
// Player start functions

function PlayStartUpMessage(PlayerPawn NewPlayer)
{
	local int i;
	local color WhiteColor;
	local AlignSPP Ali;
	local color ColorGold, ColorPink;
	local string tmp;
	local StatLog SL;

	NewPlayer.ClearProgressMessages();
	ColorGold.R = 254;
	ColorGold.G = 254;
	ColorPink.R = 254;
	ColorPink.G = 128;
	ColorPink.B = 254;

	if (MapMessage == "")
	{
		SL = Spawn(class'StatLog');
		if ( SL != None )
		{
			tmp = SL.GetMapFileName();
			SL.Disable('Timer');
			SL.Destroy();
		}
//	MapMessage = "Running Map:"@Level.GetURLMap();
	if ( tmp != "" )
		MapMessage = "Running Map:"@tmp;
	else
		MapMessage = "Bork > Unknown Map Name!!!";
	}

	if ( Level.NetMode != NM_StandAlone ) //Hack for client's TimeDilation when Level is retarded + NetSpeed
	{									//Using a setup Classic with 1.000000 that crap seems not replicated to client
		Ali = Spawn(class'AlignSPP',NewPlayer); //And client remains screwed - so we force tracking Dilation from server
		if (Ali != None)
			Ali.Speed = GameSpeed;
	}

	// GameName
	NewPlayer.SetProgressMessage(GameName, i++);
	if ( bRequireReady && (Level.NetMode != NM_Standalone) )
	{
		NewPlayer.SetProgressMessage(TourneyMessage, i++);
		NewPlayer.SetProgressColor(ColorPink, i);
		NewPlayer.SetProgressMessage( MapMessage, i++ );
	}
	else
	{
		NewPlayer.SetProgressMessage(StartUpMessage, i++);
	}

	if ( GoalTeamScore > 0 )
		NewPlayer.SetProgressMessage(int(GoalTeamScore)@GameGoal, i++);

	if ( NewPlayer.PlayerReplicationInfo.Team < 4 )
	{
		NewPlayer.SetProgressColor(class'ChallengeTeamHUD'.Default.TeamColor[NewPlayer.PlayerReplicationInfo.Team], i);
		NewPlayer.SetProgressMessage(StartupTeamMessage@Teams[NewPlayer.PlayerReplicationInfo.Team].TeamName$StartupTeamTralier, i++);
		WhiteColor.R = 255;
		WhiteColor.G = 255;
		WhiteColor.B = 255;
		NewPlayer.SetProgressColor(WhiteColor, i);
		if ( !bRatedGame )
			NewPlayer.SetProgressMessage(TeamChangeMessage, i++);
	}

	if ( Level.NetMode == NM_Standalone )
	{
		NewPlayer.SetProgressMessage(SingleWaitingMessage, i++);
		NewPlayer.SetProgressColor(ColorPink, i);
		NewPlayer.SetProgressMessage( MapMessage, i++ );
	}
	if ( bUseXC )
	{
		NewPlayer.SetProgressColor(ColorGold, i);
		NewPlayer.SetProgressMessage("Enhanced with XC_Engine", i++ );
	}
}

function AddDefaultInventory( pawn PlayerPawn )
{
	local Weapon NewWeapon;
	local Bot B;

	if ( PlayerPawn.IsA('Spectator') || (bRequireReady && (CountDown > 0)) )
		return;

	Super(GameInfo).AddDefaultInventory(PlayerPawn);
	if ( bUseTranslocator && (!bRatedGame || bRatedTranslocator) )
	{
		// Spawn Translocator.
		if( PlayerPawn.FindInventoryType(class'CTFTranslocator')==None )
		{
			newWeapon = Spawn(class'CTFTranslocator');
			if( newWeapon != None )
			{
				newWeapon.RespawnTime = 0.00;
				newWeapon.Instigator = PlayerPawn;
				newWeapon.BecomeItem();
				PlayerPawn.AddInventory(newWeapon);
				newWeapon.GiveAmmo(PlayerPawn);
				newWeapon.SetSwitchPriority(PlayerPawn);
				newWeapon.WeaponSet(PlayerPawn);
			}
		}
	}
	if( PlayerPawn.FindInventoryType(class'ImpactHammer')==None )
	{
		newWeapon = Spawn(class'BotPack.ImpactHammer');
		if( newWeapon != None )
		{
			newWeapon.RespawnTime = 0.00;
			newWeapon.Instigator = PlayerPawn;
			newWeapon.BecomeItem();
			PlayerPawn.AddInventory(newWeapon);
			newWeapon.GiveAmmo(PlayerPawn);
			newWeapon.SetSwitchPriority(PlayerPawn);
			newWeapon.WeaponSet(PlayerPawn);
		}
	}
	if ( PlayerPawn.Weapon != None && PlayerPawn.Weapon != DefaultWeapon )
	{
		PlayerPawn.Weapon.GotoState('DownWeapon');
		PlayerPawn.PendingWeapon = None;
	}
	// Spawn Automag
	// Nelsona: What AutoMag ?
	B = Bot(PlayerPawn);
	if ( B != None )
		B.bHasImpactHammer = (B.FindInventoryType(class'ImpactHammer') != None);
}

function Logout(pawn Exiting)
{
	local Projectile Pr;
	local Actor T;
	local int i;

	foreach AllActors ( class 'Projectile',Pr )
	{
		if ( Pr.Instigator == Exiting )
		{
			Pr.bReplicateInstigator = False;
			Pr.Instigator = None;
			log (Exiting.GetHumanName()@"has gone, trash"@Pr@"will be removed.");
			Pr.Destroy();
		}
	}

	foreach Exiting.RadiusActors (class'Actor',T,Exiting.CollisionRadius*2)
	{
		for (i=0;i<4;i++)
		if ( T.Touching[i] == Exiting )
		{
			Exiting.UnTouch(T);
			Exiting.SetCollision(False,False,False);
			log ("Hit >> Logout::Exiting."$Exiting.Name$" touching"@T.Name@"attempt recover...");
		}
	}

	if ( !Exiting.bIsPlayer || Exiting.PlayerReplicationInfo == None )
		return;
	Super.Logout(Exiting);
}
// rebalance teams after player changes teams or leaves
// find biggest and smallest teams.  If 2 apart, move bot from biggest to smallest
function ReBalance()
{
	local int big, small, i, bigsize, smallsize;
	local Pawn P, A;
	local Bot B;

	if ( bBalancing || (NumBots == 0) )
		return;

	big = 0;
	small = 0;
	bigsize = Teams[0].Size;
	smallsize = Teams[0].Size;
	for ( i=1; i<MaxTeams; i++ )
	{
		if ( Teams[i].Size > bigsize )
		{
			big = i;
			bigsize = Teams[i].Size;
		}
		else if ( Teams[i].Size < smallsize )
		{
			small = i;
			smallsize = Teams[i].Size;
		}
	}
	
	bBalancing = true;
	while ( bigsize - smallsize > 1 )
	{
		for ( P=Level.PawnList; P!=None; P=P.NextPawn )
			if ( P.PlayerReplicationInfo != None && P.bIsPlayer && (P.PlayerReplicationInfo.Team == big)
				&& Bot(P) != None )
			{
				B = Bot(P);
				break;
			}
		if ( B != None )
		{
			B.Health = 0;
			B.Died( None, 'Suicided', B.Location );
			bigsize--;
			smallsize++;
			ChangeTeam(B, small);
		}
		else
			Break;
	}
	bBalancing = false;

	// re-assign orders to follower bots with no leaders
	for ( P=Level.PawnList; P!=None; P=P.NextPawn )
		if ( P.PlayerReplicationInfo != None && P.bIsPlayer && Bot(P) != None && (BotReplicationInfo(P.PlayerReplicationInfo).RealOrders == 'Follow') )
		{
			if ( Pawn(Bot(P).OrderObject) != None )
				A = Pawn(Bot(P).OrderObject);
			if ( (A == None) || A.bDeleteMe || !A.bIsPlayer || ( A.PlayerReplicationInfo.Team != P.PlayerReplicationInfo.Team && (A.PlayerReplicationInfo != None && P.PlayerReplicationInfo != None)) )
			{
				Bot(P).OrderObject = None;
				SetBotOrders(Bot(P));
			}
		}
}

function NavigationPoint FindPlayerStart( Pawn Player, optional byte InTeam, optional string incomingName )
{
	local PlayerStart Dest, Candidate[32], Best;
	local float Score[16], BestScore, NextDist;
	local pawn OtherPlayer;
	local int i, num;
	local Teleporter Tel;
	local NavigationPoint N;
	local byte Team;

	if ( bStartMatch && (Player != None) && Player.IsA('TournamentPlayer') 
		&& (Level.NetMode == NM_Standalone)
		&& (TournamentPlayer(Player).StartSpot != None) )
		return TournamentPlayer(Player).StartSpot;

	if ( (Player != None) && (Player.PlayerReplicationInfo != None) )
		Team = Player.PlayerReplicationInfo.Team;
	else
		Team = InTeam;

	if( incomingName!="" )
		foreach AllActors( class 'Teleporter', Tel )
			if( string(Tel.Tag)~=incomingName )
				return Tel;

	if ( Team == 255 )
		Team = 0;

	//choose candidates
	for ( N=Level.NavigationPointList; N!=None; N=N.nextNavigationPoint )
	{
		Dest = PlayerStart(N);
		if ( (Dest != None) && Dest.bEnabled
			&& (!bSpawnInTeamArea || (Team == Dest.TeamNumber)) )
		{
			if (num<32)
				Candidate[num] = Dest;
			else if (Rand(num) < 32)
				Candidate[Rand(32)] = Dest;
			num++;
		}
	}

	if (num == 0 )
	{
//		log("Didn't find any player starts in list for team"@Team@"!!!"); 
		foreach AllActors( class'PlayerStart', Dest )
		{
			if (num<32)
				Candidate[num] = Dest;
			else if (Rand(num) < 32)
				Candidate[Rand(32)] = Dest;
			num++;
		}
		if ( num == 0 )
			return None;
	}

	if (num>32) 
		num = 32;

	//assess candidates
	for (i=0;i<num;i++)
	{
		if ( Candidate[i] == LastStartSpot )
			Score[i] = -6000.0;
		else
			Score[i] = 4000 * FRand(); //randomize
	}

	for ( OtherPlayer=Level.PawnList; OtherPlayer!=None; OtherPlayer=OtherPlayer.NextPawn)
		if (OtherPlayer.PlayerReplicationInfo != None && OtherPlayer.bIsPlayer && (OtherPlayer.Health > 0) && !OtherPlayer.IsA('Spectator') )
			for (i=0; i<num; i++)
				if ( OtherPlayer.Region.Zone == Candidate[i].Region.Zone ) 
				{
					Score[i] -= 1500;
					NextDist = VSize(OtherPlayer.Location - Candidate[i].Location);
					if (NextDist < 2 * (CollisionRadius + CollisionHeight))
						Score[i] -= 1000000.0;
					else if ( (NextDist < 2000) && (OtherPlayer.PlayerReplicationInfo.Team != Team)
							&& FastTrace(Candidate[i].Location, OtherPlayer.Location) )
						Score[i] -= (10000.0 - NextDist);
				}

	BestScore = Score[0];
	Best = Candidate[0];
	for (i=1; i<num; i++)
		if (Score[i] > BestScore)
		{
			BestScore = Score[i];
			Best = Candidate[i];
		}
	LastStartSpot = Best;
	return Best;
}

//-------------------------------------------------------------------------------------
// Level gameplay modification

//Use reduce damage for teamplay modifications, etc.
function int ReduceDamage(int Damage, name DamageType, pawn injured, pawn instigatedBy)
{
	local bool bKL, bV;

	bKL = ( instigatedBy != None );
	bV = ( injured != None );

	if ( bV )
	{
		if (injured.Region.Zone.bNeutralZone)
			return 0;
		if ( injured.bIsPlayer && injured.PlayerReplicationInfo == None )
			injured.bIsPlayer = False;
	}
	else
		Return 0;
	if ( !bKL )
		return Damage;
	if ( bKL )
		if ( InstigatedBy.bIsplayer && InstigatedBy.PlayerReplicationInfo == None)
			InstigatedBy.bIsPlayer = False;

	Damage *= InstigatedBy.DamageScaling;
	if ( bHardCoreMode )
		Damage *= 1.5;
	if ( bNoviceMode && !bThreePlus )
	{
		if ( instigatedBy.bIsPlayer && (injured == instigatedby) && (Level.NetMode == NM_Standalone) )
			Damage *= 0.5;

		//skill level modification
		if ( instigatedBy.IsA('Bot') && injured.IsA('PlayerPawn') )
		{
			if ( ((instigatedBy.Weapon != None) && instigatedBy.Weapon.bMeleeWeapon) 
				|| ((injured.Weapon != None) && injured.Weapon.bMeleeWeapon && (VSize(injured.location - instigatedBy.Location) < 600)) )
				Damage = Damage * (0.76 + 0.08 * instigatedBy.skill);
			else
				Damage = Damage * (0.25 + 0.15 * instigatedBy.skill);
		}
	}

	if ( (instigatedBy != injured) && injured.PlayerReplicationInfo != None && instigatedBy.PlayerReplicationInfo != None
		&& (injured.PlayerReplicationInfo.Team ~= instigatedBy.PlayerReplicationInfo.Team) )
	{
		if ( injured.IsA('Bot') )
			Bot(Injured).YellAt(instigatedBy);
		return (Damage * FriendlyFireScale);
	}
	return Damage;
}

function Killed(pawn killer, pawn Other, name damageType)
{
	local int NextTaunt, i;
	local bool bAutoTaunt, bEndOverTime;
	local Pawn P, Best;
	local Projectile Pr;
	// Section: Turns Off bullshits
	foreach AllActors ( class 'Projectile', Pr )
	{
		if ( Pr != None && Pr.Instigator == Other )		// I don't check for "Other" because that one called this function not "Killer"
		{
			Pr.bReplicateInstigator = False;
			Pr.Instigator = None;
			Pr.Destroy();
		}
	}
	// End of Bullshits Destruction

	if ( Killer != None && Killer.PlayerReplicationInfo == None && Other.PlayerReplicationInfo != None )
	{
		Other.DieCount++;
		Other.PlayerReplicationInfo.Deaths += 1;
		Other.PlayerReplicationInfo.Score -= 1;
//		log (Other.GetHumanName()$" died having "$Other.PlayerReplicationInfo.Deaths);
	}
	if ( (damageType == 'Decapitated') && (Killer != Other) && (Killer != None) && Killer.PlayerReplicationInfo != None )
	{
		if ( Killer.bIsHuman ) //Bot need to receive this crap, dude ? I think Loque doesn't give a shit on this :|
			Killer.ReceiveLocalizedMessage( class'DecapitationMessage' );
	}
//Moved before to call super because GameInfo is not ready for monsters and won't be soon. Sleep well, UTPG ! :|
	if ( Killer != None)
	{
		if ( Killer.PlayerReplicationInfo != None && Killer != Other )
			Killer.PlayerReplicationInfo.Score += 1;
		if ( Other.Spree > 4 && Other.PlayerReplicationInfo != None )
		{
			if ( Killer.PlayerReplicationInfo != None )
				EndSpree(Killer, Other);
		}
		Other.Spree = 0;
		if ( !bFirstBlood )
			if ( Killer != Other && Killer.PlayerReplicationInfo != None )
				if (!Self.IsA('TrainingDM'))
				{
					BroadcastLocalizedMessage( class'FirstBloodMessage', 0, Killer.PlayerReplicationInfo );
				}
		if ( /*Other.bIsPlayer End of crap &&*/ (Killer != None) && (Killer != Other) && Killer.PlayerReplicationInfo != None )
		{
			Killer.Spree++;
			if ( Killer.Spree > 4 )
				NotifySpree(Killer, Killer.Spree);
		} 
		if ( Killer.IsA('ScriptedPawn') && Killer != Other )
			ScriptedPawn(Killer).WhatToDoNext('Roaming','Roam');
		if ( bRatedGame ) //hmmmm... 
			RateVs(Other, Killer); // Is returned if is out of normality so I think is not a problem
		bFirstBlood = True;
		if ( !bTeamGame )
		{
			if ( bOverTime )
			{
				bEndOverTime = true;
				//check for clear winner now
				// find individual winner
				for ( P=Level.PawnList; P!=None; P=P.nextPawn )
				{
					if ( P.PlayerReplicationInfo == None )
						continue;
					if ( P.PlayerReplicationInfo != None && ((Best == None) || (P.PlayerReplicationInfo.Score > Best.PlayerReplicationInfo.Score)) )
						Best = P;
				}
				// check for tie
				for ( P=Level.PawnList; P!=None; P=P.nextPawn )
				{
					if ( P.PlayerReplicationInfo == None )
						continue;
					if ( P.PlayerReplicationInfo != None && (Best != P) && ( P.PlayerReplicationInfo.Score == Best.PlayerReplicationInfo.Score ) )
						bEndOverTime = False;
				}
				if ( bEndOverTime )
				{
					if ( (FragLimit > 0) && (Best.PlayerReplicationInfo.Score >= FragLimit) && Best != None )
						EndGame("fraglimit");
					else
						EndGame("timelimit");
				}
			}
			else if ( (FragLimit > 0) && (killer != None) && (killer.PlayerReplicationInfo != None)
					&& (killer.PlayerReplicationInfo.Score >= FragLimit) )
				EndGame("fraglimit");
		}
		if ( Other.PlayerReplicationInfo == None )
		{
			ScoreKill(Killer,Other);
			return;
		}
	}
	else
	{
		// Completed because of new parameters and your laziness
		if ( !Other.bIsPlayer || Other.PlayerReplicationInfo == None )
		{
			if (DamageType == 'Exploded')
				BroadcastMessage(Other.GetHumanName()@" had a strange explosion.");
			else if (DamageType == 'Burned')
				BroadcastMessage(Other.GetHumanName()@" was burned by a hostile environment.");
			else if (DamageType == 'Corroded')
				BroadcastMessage(Other.GetHumanName()@" was corroded by a hostile fluid.");
			else if (DamageType == 'Mortared')
				BroadcastMessage(Other.GetHumanName()@" was blown up by a mortar.");
			else if (DamageType == 'Drowned')
				BroadcastMessage(Other.GetHumanName()@" forgot to come up for air.");
			else if (DamageType == 'Fell')
				BroadcastMessage(Other.GetHumanName()@" crashed falling.");
			else if (DamageType == 'Eradicated')
				BroadcastMessage(Other.GetHumanName()@" has been eradicated.");
			else if (DamageType == 'Suicided')						//Can be ? Movers ? Others ?
				BroadcastMessage(Other.GetHumanName()@" suffered a self damage.");
			else if (DamageType == 'Crushed')
				BroadcastMessage(Other.GetHumanName()@" died crushed.");
			else
				BroadcastMessage(Other.GetHumanName()@" died.");
			if ( ScriptedPawn(Other) != None && ScriptedPawn(Other).CarcassType != None )
				ScriptedPawn(Other).SpawnGibbedCarcass();
			Other.Destroy(); //Done
			return;
		}
		if ( Other.PlayerReplicationInfo != None && DamageType == 'SpecialDamage')
		{
			if (SpecialDamageString == "")
				BroadcastMessage(Other.GetHumanName()@"died.");
			else
				BroadcastMessage(Other.GetHumanName()@"died"@SpecialDamageString);
			Other.DieCount++;
			Other.PlayerReplicationInfo.Score -= 1;
			Other.Spree = 0;
			return;
		}
		if (Other == None) return;
	}
	Super(GameInfo).Killed(killer, Other, damageType); //Now do call if need something
}

function ScoreKill(pawn Killer, pawn Other)
{
	local TournamentPlayer TP;

	if ( (Killer == None) || (Killer == Other) || !Other.bIsPlayer || !Killer.bIsPlayer 
		|| ( Killer.PlayerReplicationInfo != None || Other.PlayerReplicationInfo != None ) )
	{
		Other.DieCount++;
		if ( Other.PlayerReplicationInfo != None )
			Other.PlayerReplicationInfo.Score -= 1;
	else if ( Killer != None )
	{
		Killer.KillCount++;
		if ( bAltScoring && (Killer != Other) && Other.PlayerReplicationInfo != None ) //DeathMatchPlus.ScoreKill Accesed None ??? No wander...
			Other.PlayerReplicationInfo.Score -= 1;

		if ( Killer.PlayerReplicationInfo != None )
		{
			if ( TournamentPlayer(Killer) != None && Other.PlayerReplicationInfo == None )
			{
				TP = TournamentPlayer(Killer);
				if ( /*TP. */ Level.TimeSeconds - TP.LastKillTime < 3 )
				{
					TP.MultiLevel++;
					TP.ReceiveLocalizedMessage( class'MultiKillMessage', TP.MultiLevel );
				} 
				else
					TP.MultiLevel = 0;
				TP.LastKillTime = TP.Level.TimeSeconds;
			}
			if ( Other != None && (!Other.bIsPlayer || Other.PlayerReplicationInfo == None) )
				BroadcastMessage(Killer.GetHumanName()@"killed "$Other.GetHumanName());
		}

		if ( Killer.PlayerReplicationInfo == None && Other.PlayerReplicationInfo == None && Killer != Other )
			BroadcastMessage(Killer.GetHumanName()@"killed "$Other.GetHumanName());

		if ( Killer.PlayerReplicationInfo == None && Other.PlayerReplicationInfo != None )
			Other.PlayerReplicationInfo.Score -= 1;

		if ( Killer == Other && Killer.PlayerReplicationInfo == None )
			BroadcastMessage(Killer.GetHumanName()@"suffered a self deadly pain.");
	}
	}

	if ( !bScoreTeamKills )
		return;
	if ( Other.PlayerReplicationInfo != None &&
	( Killer == None || (Killer != None && Killer.PlayerReplicationInfo != None)) )
	{
		if ( (Killer != none && Killer == Other) || Killer == None )
			Teams[Other.PlayerReplicationInfo.Team].Score -= 1;
		else if ( Killer.PlayerReplicationInfo.Team != Other.PlayerReplicationInfo.Team
		&& ( Killer.PlayerReplicationInfo != None && Other.PlayerReplicationInfo != None ))
			Teams[Killer.PlayerReplicationInfo.Team].Score += 1;
		else if ( FriendlyFireScale > 0 )
		{
			Teams[Other.PlayerReplicationInfo.Team].Score -= 1;
			Killer.PlayerReplicationInfo.Score -= 1;
		}
	}
	if ( Killer != None )
		if ( (bOverTime || (GoalTeamScore > 0)) && Killer.PlayerReplicationInfo != None
			&& (Teams[killer.PlayerReplicationInfo.Team].Score >= GoalTeamScore) )
			EndGame("teamscorelimit");
	BaseMutator.ScoreKill(Killer, Other); //LOL
}

function bool ChangeTeam(Pawn Other, int NewTeam)
{
	local int i, Smallest, DesiredTeam;
	local pawn APlayer, P;
	local teaminfo SmallestTeam;

	if ( bRatedGame && (Other.PlayerReplicationInfo.Team != 255) )
		return false;
	if ( Other.IsA('Spectator') )
	{
		Other.PlayerReplicationInfo.Team = 255;
		if (LocalLog != None)
			LocalLog.LogTeamChange(Other);
		if (WorldLog != None)
			WorldLog.LogTeamChange(Other);
		return true;
	}

	// find smallest team
	Smallest = 0;
	for( i=1; i<MaxTeams; i++ )
		if ( Teams[Smallest].Size > Teams[i].Size )
			Smallest = i;

	if ( (NewTeam == 255) || (NewTeam >= MaxTeams) )
		NewTeam = Smallest;

	if ( bPlayersBalanceTeams && (Level.NetMode != NM_Standalone) )
	{
		if ( Teams[NewTeam].Size > Teams[Smallest].Size )
			NewTeam = Smallest;
		if ( NumBots == 1 )
		{
			// join bot's team if sizes are equal, because he will leave
			for ( P=Level.PawnList; P!=None; P=P.NextPawn )
				if ( P.IsA('Bot') )
					break;
			
			if ( (P != None) && (P.PlayerReplicationInfo != None) && (P.PlayerReplicationInfo.Team != 255)
				&& (Teams[P.PlayerReplicationInfo.Team].Size == Teams[Smallest].Size) )
				NewTeam = P.PlayerReplicationInfo.Team;
		}
	}

	if ( (Other.PlayerReplicationInfo.Team == NewTeam) && bNoTeamChanges )
		return false;

	if ( Other.IsA('TournamentPlayer') )
		TournamentPlayer(Other).StartSpot = None;

	if ( Other.PlayerReplicationInfo.Team != 255 )
	{
		ClearOrders(Other);
		Teams[Other.PlayerReplicationInfo.Team].Size--;
	}

	if ( Teams[NewTeam].Size < MaxTeamSize )
	{
		AddToTeam(NewTeam, Other);
		return true;
	}

	if ( Other.PlayerReplicationInfo.Team == 255 )
	{
		AddToTeam(Smallest, Other);
		return true;
	}

	return false;
}

function AddToTeam( int num, Pawn Other )
{
	local teaminfo aTeam;
	local Pawn P;
	local bool bSuccess;
	local string SkinName, FaceName;

	if ( Other == None || Other.PlayerReplicationInfo == None )
	{
		log("Added none to team!!!");
		return;
	}

	aTeam = Teams[num];

	aTeam.Size++;
	Other.PlayerReplicationInfo.Team = num;
	Other.PlayerReplicationInfo.TeamName = aTeam.TeamName;
	if (LocalLog != None)
		LocalLog.LogTeamChange(Other);
	if (WorldLog != None)
		WorldLog.LogTeamChange(Other);
	bSuccess = false;
	if ( Other.IsA('PlayerPawn') )
	{
		Other.PlayerReplicationInfo.TeamID = 0;
		PlayerPawn(Other).ClientChangeTeam(Other.PlayerReplicationInfo.Team);
	}
	else
		Other.PlayerReplicationInfo.TeamID = 1;

	while ( !bSuccess )
	{
		bSuccess = true;
		for ( P=Level.PawnList; P!=None; P=P.nextPawn )
            if ( P.PlayerReplicationInfo != None && (P != Other) 
				&& (P.PlayerReplicationInfo.Team == Other.PlayerReplicationInfo.Team) 
				&& (P.PlayerReplicationInfo.TeamId == Other.PlayerReplicationInfo.TeamId) )
				bSuccess = false;
		if ( !bSuccess )
			Other.PlayerReplicationInfo.TeamID++;
	}

	BroadcastLocalizedMessage( DMMessageClass, 3, Other.PlayerReplicationInfo, None, aTeam );

	Other.static.GetMultiSkin(Other, SkinName, FaceName);
	Other.static.SetMultiSkin(Other, SkinName, FaceName, num);

	if ( bBalanceTeams && !bRatedGame )
		ReBalance();
}

function bool CanSpectate( pawn Viewer, actor ViewTarget )
{
	if ( ViewTarget != None && ViewTarget.bIsPawn && (Pawn(ViewTarget).PlayerReplicationInfo != None)
		&& Pawn(ViewTarget).PlayerReplicationInfo.bIsSpectator )
		return false;
	if ( Viewer.PlayerReplicationInfo != None && Viewer.PlayerReplicationInfo.bIsSpectator && (Viewer.PlayerReplicationInfo.Team == 255) )
		return true;
	return ( (Pawn(ViewTarget) != None) && Pawn(ViewTarget).PlayerReplicationInfo != None 
		&& (Pawn(ViewTarget).PlayerReplicationInfo.Team == Viewer.PlayerReplicationInfo.Team) );
}

function TeamInfo GetTeam(int TeamNum )
{
	if ( TeamNum < ArrayCount(Teams) )
		return Teams[TeamNum];
	else return None;
}

function bool IsOnTeam(Pawn Other, int TeamNum)
{
	if ( Other.PlayerReplicationInfo != None && Other.PlayerReplicationInfo.Team == TeamNum )
		return true;

	return false;
}

function bool AddBot()
{
	local bot NewBot;
	local NavigationPoint StartSpot, OldStartSpot;
	local int DesiredTeam, i, MinSize;

	NewBot = SpawnBot(StartSpot);
	if ( NewBot == None )
	{
		log("Failed to spawn bot");
		return false;
	}

	if ( bBalanceTeams && !bRatedGame )
	{
		MinSize = Teams[0].Size;
		DesiredTeam = 0;
		for ( i=1; i<MaxTeams; i++ )
			if ( Teams[i].Size < MinSize )
			{
				MinSize = Teams[i].Size;
				DesiredTeam = i;
			}
	}
	else
		DesiredTeam = NewBot.PlayerReplicationInfo.Team;
	NewBot.PlayerReplicationInfo.Team = 255;
	if ( (DesiredTeam == 255) || !ChangeTeam(NewBot, DesiredTeam) )
	{
		ChangeTeam(NewBot, NextBotTeam);
		NextBotTeam++;
		if ( NextBotTeam >= MaxTeams )
			NextBotTeam = 0;
	}

	if ( bSpawnInTeamArea )
	{
		OldStartSpot = StartSpot;
		StartSpot = FindPlayerStart(NewBot,255);
		if ( StartSpot != None )
		{
			NewBot.SetLocation(StartSpot.Location);
			NewBot.SetRotation(StartSpot.Rotation);
			NewBot.ViewRotation = StartSpot.Rotation;
			NewBot.SetRotation(NewBot.Rotation);
			StartSpot.PlayTeleportEffect( NewBot, true );
		}
		else
			StartSpot = OldStartSpot;
	}

	StartSpot.PlayTeleportEffect(NewBot, true);

	SetBotOrders(NewBot);

	// Log it.
	if (LocalLog != None)
	{
		LocalLog.LogPlayerConnect(NewBot);
		LocalLog.FlushLog();
	}
	if (WorldLog != None)
	{
		WorldLog.LogPlayerConnect(NewBot);
		WorldLog.FlushLog();
	}

	return true;
}

function SetBotOrders(Bot NewBot)
{
	local Pawn P, L;
	local int num, total;

	// only follow players, if there are any
	if ( (NumSupportingPlayer == 0)
		 || (NumSupportingPlayer < Teams[NewBot.PlayerReplicationInfo.Team].Size/2 - 1) ) 
	{
		For ( P=Level.PawnList; P!=None; P= P.NextPawn )
			if (P != None && PlayerPawn(P) != None && ( P.PlayerReplicationInfo.Team == NewBot.PlayerReplicationInfo.Team )
				&& !P.IsA('Spectator') )
		{
			num++;
			if ( (L == None) || (FRand() < 1.0/float(num)) )
				L = P;
		}

		if ( L != None )
		{
			NumSupportingPlayer++;
			NewBot.SetOrders('Follow',L,true);
//			NewBot.SetOrders('Freelance',None,True);
			return;
		}
	}
	num = 0;
	For ( P=Level.PawnList; P!=None; P= P.NextPawn )
		if ( P.PlayerReplicationInfo != None && (P.PlayerReplicationInfo.Team == NewBot.PlayerReplicationInfo.Team) )
		{
			total++;
			if ( (P != NewBot) && P.IsA('Bot') && (Bot(P).Orders == 'FreeLance') )
			{
				num++;
				if ( (L == None) || (FRand() < 1/float(num)) )
					L = P;
			}
		}

	if ( (L != None) && (FRand() < float(num)/float(total)) )
	{
//		NewBot.SetOrders('Follow',L,true);
		NewBot.SetOrders('Freelance',None,True);
		return;
	}
	NewBot.SetOrders('Freelance', None,true);
}

function byte AssessBotAttitude(Bot aBot, Pawn Other)
{
	if ( (Other.PlayerReplicationInfo != None && (aBot.PlayerReplicationInfo.Team == Other.PlayerReplicationInfo.Team))
		|| (Other.IsA('TeamCannon') 
			&& (StationaryPawn(Other).SameTeamAs(aBot.PlayerReplicationInfo.Team))) ) 
		return 3;
	else if ( Other.PlayerReplicationInfo == None && Other.Enemy != None && Other.Enemy == aBot )
		return 1;
	else 
		return Super(DeathMatchPlus).AssessBotAttitude(aBot, Other);
}

function Actor SetDefenseFor(Bot aBot)
{
	return None;
}

function bool FindSpecialAttractionFor(Bot aBot)
{
	return false;
}

function SetAttractionStateFor(Bot aBot)
{
	if ( aBot.Enemy != None )
	{
		if ( !aBot.IsInState('FallBack') )
		{
			aBot.bNoClearSpecial = true;
			aBot.TweenToRunning(0.1);
			aBot.GotoState('FallBack','SpecialNavig');
		}
	}
	else if ( !aBot.IsInState('Roaming') )
	{
		aBot.bNoClearSpecial = true;
		aBot.TweenToRunning(0.1);
		aBot.GotoState('Roaming', 'SpecialNavig');
	}
}

function PickAmbushSpotFor(Bot aBot)
{
	local NavigationPoint N;

	for ( N=Level.NavigationPointList; N!=None; N=N.NextNavigationPoint )
		if ( N.IsA('Ambushpoint') && !N.taken )
		{
			if ( aBot.Orders == 'Defend' )
			{
				if ( N.IsA('DefensePoint') && (DefensePoint(N).team == aBot.PlayerReplicationInfo.team) )
				{
					if ( (DefensePoint(aBot.Ambushspot) == None)
						|| (DefensePoint(N).priority > DefensePoint(aBot.Ambushspot).priority) )
						aBot.Ambushspot = Ambushpoint(N);
					else if ( (DefensePoint(N).priority == DefensePoint(aBot.Ambushspot).priority)
						&& (FRand() < 0.4) ) 
						aBot.Ambushspot = Ambushpoint(N);
				}
				else if ( (DefensePoint(aBot.AmbushSpot) == None)
						&& (VSize(N.Location - aBot.OrderObject.Location) < 1500)
						&& FastTrace(aBot.OrderObject.Location, N.Location)
						&& ((aBot.Ambushspot == None) || (FRand() < 0.5)) )
							aBot.Ambushspot = Ambushpoint(N);
			}
			else if ( (aBot.AmbushSpot == None)
				|| (VSize(aBot.Location - aBot.Ambushspot.Location)
					 > VSize(aBot.Location - N.Location)) )
				aBot.Ambushspot = Ambushpoint(N);
		}
}

function byte PriorityObjective(Bot aBot)
{
	return 0;
}

function bool SuccessfulGame()
{
	local TeamInfo BestTeam;
	local int i;
	BestTeam = Teams[0];
	for ( i=1; i<MaxTeams; i++ )
		if ( Teams[i].Score > BestTeam.Score )
			BestTeam = Teams[i];

	bFulfilledSpecial = True; // Override and implement if you have a special condition.
	if (BestTeam.TeamIndex == RatedPlayer.PlayerReplicationInfo.Team)
		return ( bFulfilledSpecial && (BestTeam.Score >= GoalTeamScore) );
	else
		return false;
}

function ClearOrders(Pawn Leaving)
{
	local Pawn P;

	for ( P=Level.PawnList; P!=None; P=P.NextPawn )
		if ( P.IsA('Bot') && ( Bot(P).OrderObject != None && Bot(P).OrderObject == Leaving) )
			Bot(P).SetOrders('Freelance', None);
}
function bool WaitForPoint(bot aBot)
{
	return false;
}

function bool SendBotToGoal(Bot aBot)
{
	return false;
}

function bool HandleTieUp(Bot Bumper, Bot Bumpee)
{
	return false;
}

//------------------------------------------------------------------------------
// Game Querying.

function string GetRules()
{
	local string ResultSet;
	ResultSet = Super(TournamentGameInfo).GetRules();

	ResultSet = ResultSet$"\\timelimit\\"$TimeLimit;
	ResultSet = ResultSet$"\\goalteamscore\\"$int(GoalTeamScore);
	Resultset = ResultSet$"\\minplayers\\"$MinPlayers;
	Resultset = ResultSet$"\\changelevels\\"$bChangeLevels;
	ResultSet = ResultSet$"\\maxteams\\"$MaxTeams;
	ResultSet = ResultSet$"\\balanceteams\\"$bBalanceTeams;
	ResultSet = ResultSet$"\\playersbalanceteams\\"$bPlayersBalanceTeams;
	ResultSet = ResultSet$"\\friendlyfire\\"$int(FriendlyFireScale*100)$"%";
	Resultset = ResultSet$"\\tournament\\"$bTournament;
	if(bMegaSpeed)
		Resultset = ResultSet$"\\gamestyle\\Turbo";
	else
	if(bHardcoreMode)
		Resultset = ResultSet$"\\gamestyle\\Hardcore";
	else
		Resultset = ResultSet$"\\gamestyle\\Classic";

	if(MinPlayers > 0)
		Resultset = ResultSet$"\\botskill\\"$class'ChallengeBotInfo'.default.Skills[Difficulty];

	return ResultSet;
}

function DiscardInventory( Pawn Other ) //Guess why I did a... small look around
{
	local actor dropped;
	local inventory Inv;
	local weapon weap;
	local float speed;

	CaptureActor(CaptActor).bHaveASpawn=True; //This has been called - look out how we replace stuff dropped in Base_Mute
	if( Other.DropWhenKilled != None )
	{
		dropped = Spawn(Other.DropWhenKilled,,,Other.Location); //Break in here if replacement works - next lines won't act anymore
		Inv = Inventory(dropped);
		if ( Inv != None )
		{
			Inv.RespawnTime = 0.0; //don't respawn - Nelsona: For sure not even replacements - except from decorations... RIGHT ???
			Inv.BecomePickup();
		}
		if ( dropped != None )
		{
			dropped.RemoteRole = ROLE_DumbProxy;
			dropped.SetPhysics(PHYS_Falling);
			dropped.bCollideWorld = true;
			dropped.Velocity = Other.Velocity + VRand() * 280;
		}
		if ( Inv != None )
			Inv.GotoState('PickUp', 'Dropped');
	}
	if( (Other.Weapon!=None) && (Other.Weapon.Class!=Level.Game.BaseMutator.MutatedDefaultWeapon()) 
		&& Other.Weapon.bCanThrow )
	{
		speed = VSize(Other.Velocity);
		weap = Other.Weapon;
		if (speed != 0)
			weap.Velocity = Normal(Other.Velocity/speed + 0.5 * VRand()) * (speed + 280);
		else {
			weap.Velocity.X = 0;
			weap.Velocity.Y = 0;
			weap.Velocity.Z = 0;
		}
		Other.TossWeapon();
		if ( weap.PickupAmmoCount == 0 )
			weap.PickupAmmoCount = 1;
	}
	Other.Weapon = None;
	Other.SelectedItem = None;
	for( Inv=Other.Inventory; Inv!=None; Inv=Inv.Inventory )
		Inv.Destroy();
	CaptureActor(CaptActor).bHaveASpawn=False; //End, prepare to next one
}

defaultproperties
{
   bScoreTeamKills=True
   bBalanceTeams=True
   bPlayersBalanceTeams=True
   MaxTeams=2
   MaxAllowedTeams=4
   GoalTeamScore=30.000000
   MaxTeamSize=16
   StartUpTeamMessage="You are on"
   TeamChangeMessage="Use Options->Player Setup to change teams."
   TeamColor(0)="Red"
   TeamColor(1)="Blue"
   TeamColor(2)="Green"
   TeamColor(3)="Gold"
   TEAM_Blue=1
   TEAM_Green=2
   TEAM_Gold=3
   CurrentOrders(0)=Defend
   CurrentOrders(1)=Defend
   CurrentOrders(2)=Defend
   CurrentOrders(3)=Defend
   StartupTeamTralier="."
   TimeLimit=10
   bUseTranslocator=False
   StartUpMessage="Work with your teammates against the other teams."
   bCanChangeSkin=False
   bTeamGame=True
   ScoreBoardType=Class'Botpack.TeamScoreBoard'
   RulesMenuType="UTMenu.UTTeamRSClient"
   SettingsMenuType="UTMenu.UTTeamSSClient"
   HUDType=Class'Botpack.ChallengeTeamHUD'
   BeaconName="TTeam"
   ExplodeMessage=" was directly blown up"
   FallMessage=" left just a small crater."
   DrownedMessage=" has drowned because of missing air."
   BurnedMessage=" was so incinerated"
   CorrodedMessage=" was nasty slimed"
   HackedMessage=" was successfully hacked"
   MortarMessage=" was blasted by a mortar."
   GameName="Ns Tournament Team Game"
   MutatorClass=Class'NsTDM2.NsDeathMatchMutator'
   StatLogClass=Class'NsTDM2.DMLogFile'
   DefaultWeapon=class'UnrealShare.DispersionPistol'
}
